from __future__ import annotations

import os
import shutil
import tempfile

from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc


_shims = ("node", "npm", "npx", "yarn", "pnpm")


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")
        top_level_dir = "node-env"
        archive.unpack(
            archive_file,
            tmpd,
            perform_strip1=True,
            strip1_new_prefix=top_level_dir,
        )

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/{top_level_dir}", f"{into}/{top_level_dir}")


def uninstall(binroot: str) -> None:
    shutil.rmtree(f"{binroot}/node-env", ignore_errors=True)

    for shim in (*_shims, "yarn"):
        fp = f"{binroot}/{shim}"
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def installed(version: str, binroot: str) -> bool:
    if shutil.which(
        "node", path=binroot
    ) != f"{binroot}/node" or not os.path.exists(
        f"{binroot}/node-env/bin/node"
    ):
        return False

    with open(f"{binroot}/node", "r") as f:
        sample = f.read(64)
        if "VOLTA_HOME" in sample:
            print("volta-based node detected")
            return False

    stdout = proc.run((f"{binroot}/node", "--version"), stdout=True)
    installed_version = stdout.strip()
    if version == installed_version:
        return True

    print(f"installed node {installed_version} is outdated!")
    return False


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    binroot = fs.ensure_binroot(reporoot)

    if installed(version, binroot):
        return

    print(f"installing node {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    for shim in _shims:
        fs.write_script(
            f"{binroot}/{shim}",
            f"""#!/bin/sh
exec /usr/bin/env PATH={binroot}/node-env/bin:"$PATH" {binroot}/node-env/bin/{shim} "$@"
""",
        )

    if not installed(version, binroot):
        raise SystemExit(f"failed to install node {version}!")


def installed_yarn(version: str, binroot: str) -> bool:
    if shutil.which(
        "yarn", path=binroot
    ) != f"{binroot}/yarn" or not os.path.exists(
        f"{binroot}/node-env/bin/yarn"
    ):
        return False

    with open(f"{binroot}/yarn", "r") as f:
        sample = f.read(64)
        if "VOLTA_HOME" in sample:
            print("volta-based yarn detected")
            return False

    stdout = proc.run((f"{binroot}/yarn", "--version"), stdout=True)
    installed_version = stdout.strip()
    if version == installed_version:
        return True

    print(f"installed yarn {installed_version} is outdated!")
    return False


def install_yarn(version: str, reporoot: str) -> None:
    binroot = fs.ensure_binroot(reporoot)

    if installed_yarn(version, binroot):
        return

    # not explicitly uninstalling here, following steps
    # will pave over it
    print(f"installing yarn {version}...")

    proc.run((f"{binroot}/npm", "install", "-g", f"yarn@{version}"))

    fs.write_script(
        f"{binroot}/yarn",
        f"""#!/bin/sh
exec /usr/bin/env PATH={binroot}/node-env/bin:"$PATH" {binroot}/node-env/bin/yarn "$@"
""",
    )

    if not installed_yarn(version, binroot):
        raise SystemExit(f"failed to install yarn {version}!")



from __future__ import annotations

import io
import os
import pathlib
import tarfile
import time
import typing
import urllib.request
from unittest import mock

import pytest

from devenv.lib import archive


@pytest.fixture
def tar(tmp_path: pathlib.Path) -> pathlib.Path:
    plain = tmp_path.joinpath("plain")
    plain.write_text("hello world\n")

    tar = tmp_path.joinpath("tar")

    with tarfile.open(tar, "w:tar") as tarf:
        tarf.add(plain, arcname="hello.txt")

    return tar


@pytest.fixture
def tgz(tmp_path: pathlib.Path) -> pathlib.Path:
    a = tmp_path.joinpath("a")
    a.write_text("a")
    b = tmp_path.joinpath("b")
    b.write_text("b")

    tar = tmp_path.joinpath("tgz")

    with tarfile.open(tar, "w:gz") as tarf:
        # faster to arcname than to actually mkdirs in tmp_path
        tarf.add(a, arcname="foo-v1/bin/foo")
        tarf.add(b, arcname="foo-v1/baz")

    return tar


@pytest.fixture
def mock_sleep() -> typing.Generator[mock.MagicMock, None, None]:
    with mock.patch.object(time, "sleep", autospec=True) as mock_sleep:
        yield mock_sleep


def test_download(tmp_path: pathlib.Path, mock_sleep: mock.MagicMock) -> None:
    data = b"foo\n"
    data_sha256 = (
        "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c"
    )

    err = urllib.error.HTTPError(
        "https://example.com/foo",
        503,
        "Service Unavailable",
        "",  # type: ignore
        io.BytesIO(b""),
    )

    dest = f"{tmp_path}/a"
    with mock.patch.object(
        urllib.request,
        "urlopen",
        autospec=True,
        side_effect=(err, err, err, io.BytesIO(data)),
    ):
        archive.download("https://example.com/foo", data_sha256, dest)

    # successful download after 3 retries
    assert mock_sleep.mock_calls == [
        mock.call(1.0),
        mock.call(2.0),
        mock.call(4.0),
    ]
    with open(dest, "rb") as f:
        assert f.read() == data

    dest = f"{tmp_path}/b"
    with pytest.raises(RuntimeError) as excinfo:
        with mock.patch.object(
            urllib.request,
            "urlopen",
            autospec=True,
            side_effect=(err, err, err, err),
        ):
            archive.download("https://example.com/foo", data_sha256, dest)

    # exceeded retry limits
    assert (
        f"{excinfo.value}"
        == "Error getting https://example.com/foo: HTTP Error 503: Service Unavailable"
    )

    dest = f"{tmp_path}/b"
    with pytest.raises(RuntimeError) as excinfo:
        with mock.patch.object(
            urllib.request,
            "urlopen",
            autospec=True,
            side_effect=(io.BytesIO(data),),
        ):
            archive.download("https://example.com/foo", "wrong sha", dest)

    # sha mismatch
    assert (
        f"{excinfo.value}"
        == f"checksum mismatch for https://example.com/foo:\n- got: {data_sha256}\n- expected: wrong sha\n"
    )


def test_unpack_tar(tar: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path.joinpath("dest")
    archive.unpack(str(tar), str(dest))
    assert dest.joinpath("hello.txt").read_text() == "hello world\n"


def test_unpack_tgz_strip1(tgz: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path.joinpath("dest")
    archive.unpack(str(tgz), str(dest), perform_strip1=True)
    assert os.path.exists(f"{tmp_path}/dest/bin/foo")
    assert os.path.exists(f"{tmp_path}/dest/baz")

    dest2 = tmp_path.joinpath("dest2")
    archive.unpack(
        str(tgz), str(dest2), perform_strip1=True, strip1_new_prefix="node"
    )
    assert os.path.exists(f"{tmp_path}/dest2/node/bin/foo")
    assert os.path.exists(f"{tmp_path}/dest2/node/baz")



from __future__ import annotations

import hashlib
import os
import secrets
import shutil
import tarfile
import tempfile
import time
import urllib.request
from collections.abc import Sequence
from urllib.error import HTTPError

from devenv.constants import home


def atomic_replace(src: str, dest: str) -> None:
    if os.path.dirname(src) != os.path.dirname(dest):
        raise RuntimeError(
            f"cannot atomically move to dest {dest}; it needs to be in the same dir as {src}"
        )
    os.replace(src, dest)


def download(
    url: str,
    sha256: str,
    dest: str = "",
    retries: int = 3,
    retry_exp: float = 2.0,
) -> str:
    if retries < 0:
        raise ValueError("Retries cannot be negative")

    if not dest:
        cache_root = f"{home}/.cache/sentry-devenv"
        dest = f"{cache_root}/{sha256}"
        os.makedirs(cache_root, exist_ok=True)

    if not os.path.exists(dest):
        retry_sleep = 1.0
        while retries >= 0:
            try:
                resp = urllib.request.urlopen(url)
                break
            except HTTPError as e:
                if retries == 0:
                    raise RuntimeError(f"Error getting {url}: {e}")
                print(f"Error getting {url} ({retries} retries left): {e}")

            time.sleep(retry_sleep)
            retries -= 1
            retry_sleep *= retry_exp

        dest_dir = os.path.dirname(dest)
        os.makedirs(dest_dir, exist_ok=True)

        with tempfile.NamedTemporaryFile(delete=False, dir=dest_dir) as tmpf:
            shutil.copyfileobj(resp, tmpf)
            tmpf.seek(0)
            checksum = hashlib.sha256()
            buf = tmpf.read(4096)
            while buf:
                checksum.update(buf)
                buf = tmpf.read(4096)

            if not secrets.compare_digest(checksum.hexdigest(), sha256):
                raise RuntimeError(
                    f"checksum mismatch for {url}:\n"
                    f"- got: {checksum.hexdigest()}\n"
                    f"- expected: {sha256}\n"
                )

            atomic_replace(tmpf.name, dest)

    return dest


# mutates members!
# strips the leading component (/ is always stripped and doesn't count)
# and optionally replaces with a new prefix
def strip1(members: Sequence[tarfile.TarInfo], new_prefix: str = "") -> None:
    for member in members:
        i = member.path.find("/")
        if i == -1:
            continue
        elif i == 0:
            i = member.path[1:].find("/") + 1
            if i == 0:
                continue

        member.path = member.path[i + 1 :]  # noqa: E203

        if new_prefix:
            member.path = f"{new_prefix}/{member.path}"


def unpack(
    path: str,
    into: str,
    perform_strip1: bool = False,
    strip1_new_prefix: str = "",
) -> None:
    os.makedirs(into, exist_ok=True)
    with tarfile.open(name=path, mode="r:*") as tarf:
        if perform_strip1:
            strip1(tarf.getmembers(), strip1_new_prefix)
        tarf.extractall(into, filter="tar")
